# -*- coding: iso-8859-1 -*-
"""
Bigramme und Schtzer.  

Vgl. Skript: "Der Good-Turing-Schtzer III"
"""

import operator, goodturing

try:
	a = sum(range(2))
except NameError:
	def sum(seq):
		return reduce(operator.add, seq, 0)


class Error(Exception):
	pass


class BigramModel:
	"""This class is an abstract base class for bigram models, i.e.,
	you can feed data (which is aggregated in the probs attribute),
	compute an estimation for the distribution underlying your data
	(which is kept in probs when available) and retrieve estimates.
	"""
	def __init__(self):
		self.observation = {}
		self.probs = None
	
	def feed(self, item1, item2, count):
		"""enters count as the frequency of (item1, item2), overwriting
		any previous values.
		"""
		self.observation[item1, item2] = count
		self.probs = None
	
	def estimate(self):
		"""computes the estimation.  Has to be called before getProb can
		be used.
		"""
		raise Error, "Abstract BigramModel can't estimate"
	
	def getProb(self, item1, item2):
		"""returns the probability of the bigram (item1, item2).
		"""
		raise Error, "Abstract BigramModel can't compute probabilities"

	def computeSeqProb(self, seq):
		"""computes the probability of seq under a Markov assumption.
		"""
		return reduce(operator.mul, [self.getProb(lead, trail)
			for lead, trail in zip(seq[:-1], seq[1:])])


class MaximumLikelihoodBigramModel(BigramModel):
	"""A Bigram Model that estimates with Maximum Likelihood.
	"""
	def estimate(self):
		total = float(sum(self.observation.values()))
		self.probs = dict([(key, self.observation[key]/total)
			for key in self.observation])
	
	def getProb(self, item1, item2):
		if self.probs is None:
			self.estimate()
		return self.probs.get((item1, item2), 0.)


class GoodTuringBigramModel(BigramModel):
	"""A Bigram Model that is based on Simplified Good Turing.
	"""
	def __init__(self, *args):
		BigramModel.__init__(self, *args)
		self.seenItems = {}

	def estimate(self):
		"""In addition to computing the probs attribute as advertised in
		BigramModel, we also compute unseenProb, the probability for any
		bigram we have not seen.  We do this by evenly distributing the
		unseenMass on all unseen Species.
		"""
		unseenMass, self.probs = goodturing.getGoodTuringEstimate(
			self.observation)
		numSeenSpecies = len(self.probs)
		numPossibleSpecies = len(self.seenItems)**2
		self.unseenProb = unseenMass/(numPossibleSpecies-numSeenSpecies)
	
	def feed(self, item1, item2, count):
		"""We override this from BigramModel because we want to keep track
		how many different items there are.  We assume that everything that
		could be in item2 occurs in item1 as well.
		"""
		self.seenItems[item1] = None
		return BigramModel.feed(self, item1, item2, count)

	def getProb(self, item1, item2):
		if self.probs is None:
			self.estimate()
		return self.probs.get((item1, item2), self.unseenProb)


def _setupModels(sourceFile):
	"""just a helper for runTest.
	"""
	mlEst = MaximumLikelihoodBigramModel()
	gtEst = GoodTuringBigramModel()
	for ln in open(sourceFile):
		feedArgs = ln[0], ln[1], int(ln[3:])
		mlEst.feed(*feedArgs)
		gtEst.feed(*feedArgs)
	mlEst.estimate()
	gtEst.estimate()
	return mlEst, gtEst


def runTest(sourceFile):
	"""sets up both an ML and a SGT estimator for bigram probabilities
	using data from sourceFile (which has to have the format
	<char1><char2><blank><absolute frequency>) and then asks
	for input.  The input should consist of two words, once spelt
	correctly, once misspelt.  The program then computes which of
	the two is more likely according to both of the estimators.
	"""
	mlEst, gtEst = _setupModels(sourceFile)
	while 1:
		words = raw_input("Ein Wort, einmal richtig, einmal vertippt? ").split()
		if not len(words)==2:
			break
		good, bad = words
		if len(good)!=len(bad):
			print "Wrter mssen die gleiche Lnge haben -- sonst ist es unfair"
			continue
		print "Maximum Likelihood %g %g"%(mlEst.computeSeqProb(good),
			mlEst.computeSeqProb(bad))
		print "Simplified GT %g %g"%(gtEst.computeSeqProb(good),
			gtEst.computeSeqProb(bad))


if __name__=="__main__":
	runTest("bigrams.dist")
